package room

import (
	"cs"
	"encoding/json"
	"f5"
	"github.com/golang/protobuf/proto"
	"main/common"
	"main/constant"
	. "main/global"
	"q5"
)

type roomMgr struct {
	cs.MsgHandlerImpl
	currRoomId   int32
	idHash       map[string]*room
	roomList     q5.ListHead
}

func (this *roomMgr) Init() {
	this.idHash = make(map[string]*room)
	this.roomList.Init(nil)
	this.currRoomId = 10000
}

func (this *roomMgr) UnInit() {
}

func (this *roomMgr) getRoom(roomId string) *room {
	room, ok := this.idHash[roomId]
	if ok {
		return room
	}
	return nil
}

func (this *roomMgr) genRoomId() int32 {
	this.currRoomId++
	return this.currRoomId
}

func (this *roomMgr) genRoomIdx() int64 {
	return f5.GetApp().NewNodeUuid()
}

func (this *roomMgr) ProcessCMMsg(handler *cs.CsNetMsgHandler, hdr *f5.MsgHdr) {
	switch handler.HandlerId {
	case constant.ROOM_MGR_HANDLER_ID:
		hum := GetPlayerMgr().GetPlayerBySocket(hdr.GetSocket())
		if hum != nil {
			hdr.Context = hum
			cs.DispatchMsg(handler, hdr, this)
		}
	case constant.ROOM_HANDLER_ID:
		hum := GetPlayerMgr().GetPlayerBySocket(hdr.GetSocket())
		if hum != nil && hum.GetRoom() != nil {
			hdr.Context = hum
			cs.DispatchMsg(handler, hdr, hum.GetRoom().(*room))
		}
	}
}

func (this *roomMgr) CMCreateRoom(hdr *f5.MsgHdr, msg *cs.CMCreateRoom) {
	hum := hdr.Context.(common.Player)
	rspMsg := &cs.SMCreateRoom{}
	if hum.GetRoom() != nil && hum.GetRoom().GetRoomState() == ROOM_INIT_STATE {
		rspMsg.Errcode = proto.Int32(1)
		rspMsg.Errmsg = proto.String("already room")
		hum.SendMsg(rspMsg)
		return
	}
	m := newRoom()
	m.init(q5.ToString(this.genRoomId()), this.genRoomIdx(), hum, msg)
	this.idHash[m.roomId] = m
	this.roomList.AddTail(&m.entry)
	rspMsg.RoomId = proto.String(m.roomId)

	rspMsg.Room = new(cs.MFRoom)
	m.fillMFRoom(hum, rspMsg.Room)

	teamInfo := &TeamInfo{}
	err := json.Unmarshal([]byte(msg.GetTeamInfo()), teamInfo)
	if err == nil {
		t := m.getTeamByUuid(msg.GetTeamUuid())
		if t != nil {
			t.SaveTeamLeader(teamInfo)
		}
	}

	hum.SendMsg(rspMsg)
}

func (this *roomMgr) CMJoinRoom(hdr *f5.MsgHdr, msg *cs.CMJoinRoom) {
	hum := hdr.Context.(common.Player)
	rspMsg := cs.SMJoinRoom{}
	if hum.GetRoom() != nil && hum.GetRoom().GetRoomState() == ROOM_INIT_STATE {
		rspMsg.Errcode = proto.Int32(1)
		rspMsg.Errmsg = proto.String("already room")
		hum.SendMsg(&rspMsg)
		return
	}
	if msg.GetTeamUuid() == "" {
		rspMsg.Errcode = proto.Int32(4)
		rspMsg.Errmsg = proto.String("team_uuid is empty")
		hum.SendMsg(&rspMsg)
		return
	}
	r := this.getRoom(msg.GetRoomId())
	if r == nil {
		rspMsg.Errcode = proto.Int32(2)
		rspMsg.Errmsg = proto.String("room not found")
		hum.SendMsg(&rspMsg)
		return
	}
	if !r.canJoin(hum, msg) {
		rspMsg.Errcode = proto.Int32(3)
		rspMsg.Errmsg = proto.String("cant join")
		hum.SendMsg(&rspMsg)
		return
	}

	r.join(hum, msg)
	r.notifyRoomInfo(hum)

	/*
	if len(msg.GetTeamInfo()) > 0 {
		teamInfo := &TeamInfo{}
		err := json.Unmarshal([]byte(msg.GetTeamInfo()), teamInfo)
		if err == nil {
			t := r.getTeamByUuid(msg.GetTeamUuid())
			t.SaveTeamLeader(teamInfo)
		}
	}*/

	rspMsg.Room = new(cs.MFRoom)
	r.fillMFRoom(hum, rspMsg.Room)
	hum.SendMsg(&rspMsg)
}

func (this *roomMgr) CMSearchRoom(hdr *f5.MsgHdr, msg *cs.CMSearchRoom) {
	hum := hdr.Context.(common.Player)
	rspMsg := cs.SMSearchRoom{}
	rspMsg.Rows = make([]*cs.MFRoom, 0, constant.SEARCH_ROOM_PAGE_SIZE)
	sinceId := msg.GetSinceId()
	roomId := msg.GetRoomId()
	this.roomList.ForEach(
		func(data interface{}) bool {
			r := data.(*room)
			if r.roomIdx > sinceId && r.viewable() {
				if roomId != "" && roomId != r.roomId {
					return true
				}
				pb := new(cs.MFRoom)
				r.fillMFRoom(hum, pb)
				rspMsg.Rows = append(rspMsg.Rows, pb)
				sinceId = r.roomIdx
				if len(rspMsg.Rows) >= constant.SEARCH_ROOM_PAGE_SIZE {
					return false
				}
			}
			return true
		})
	rspMsg.SinceId = proto.Int64(sinceId)
	hum.SendMsg(&rspMsg)
}

func (this *roomMgr) CMGetCurrentRoom(hdr *f5.MsgHdr, msg *cs.CMGetCurrentRoom) {
	hum := hdr.Context.(common.Player)
	rspMsg := &cs.SMGetCurrentRoom{}
	if hum.GetRoom() != nil && hum.GetRoom().GetRoomState() == ROOM_INIT_STATE {
		rspMsg.Room = new(cs.MFCurrentRoom)
		hum.GetRoom().(*room).fillMFCurrentRoom(hum, rspMsg.Room)
		hum.SendMsg(rspMsg)
		return
	}
	hum.SendMsg(rspMsg)
}

func (this *roomMgr) RemoveRoomMember(hdr *f5.MsgHdr) {
	hum := GetPlayerMgr().GetPlayerBySocket(hdr.GetSocket())
	if hum != nil {
		if roomPtr, ok := hum.GetRoom().(*room); ok {
			roomPtr.removeMember(hum.GetAccountId())
			if len(roomPtr.teamUuidHash) <= 0 {
				roomPtr.unInit()
			}
		}
	}
}

func (this *roomMgr) CMEnterObserver(hdr *f5.MsgHdr, msg *cs.CMEnterObserver) {
	hum := hdr.Context.(common.Player)
	rspMsg := cs.SMEnterObserver{}

	roomPtr, ok := hum.GetRoom().(*room)
	if !ok || roomPtr == nil {
		rspMsg.Errcode = proto.Int32(1)
		rspMsg.Errmsg = proto.String("room is empty")
		hum.SendMsg(&rspMsg)
		return
	}

	memberPtr := roomPtr.getMember(hum.GetAccountId())
	//if memberPtr == nil || memberPtr.isLeader == 0 {
	if memberPtr == nil {
		rspMsg.Errcode = proto.Int32(2)
		rspMsg.Errmsg = proto.String("not team leader")
		hum.SendMsg(&rspMsg)
		return
	}

	if memberPtr.team.isObserver() {
		hum.SendMsg(&rspMsg)
		roomPtr.notifyRoomInfo(hum)
		return;
	}

	if roomPtr.isObTeamFull() {
		rspMsg.Errcode = proto.Int32(3)
		rspMsg.Errmsg = proto.String("ob team is full")
		hum.SendMsg(&rspMsg)
		return
	}

	t := memberPtr.team
	roomPtr.removeTeam(t)
	roomPtr.addObTeam(t)

	hum.SendMsg(&rspMsg)
	roomPtr.notifyRoomInfo(hum)
}

func (this *roomMgr) CMLeaveObserver(hdr *f5.MsgHdr, msg *cs.CMLeaveObserver) {
	hum := hdr.Context.(common.Player)
	rspMsg := cs.SMLeaveObserver{}

	roomPtr, ok := hum.GetRoom().(*room)
	if !ok || roomPtr == nil {
		rspMsg.Errcode = proto.Int32(1)
		rspMsg.Errmsg = proto.String("room is empty")
		hum.SendMsg(&rspMsg)
		return
	}

	memberPtr := roomPtr.getMember(hum.GetAccountId())
	//if memberPtr == nil || memberPtr.isLeader == 0 {
	if memberPtr == nil {
		rspMsg.Errcode = proto.Int32(2)
		rspMsg.Errmsg = proto.String("not team leader")
		hum.SendMsg(&rspMsg)
		return
	}

	if !memberPtr.team.isObserver() {
		hum.SendMsg(&rspMsg)
		roomPtr.notifyRoomInfo(hum)
		return;
	}

	if roomPtr.isTeamFull() {
		rspMsg.Errcode = proto.Int32(3)
		rspMsg.Errmsg = proto.String("room team is full")
		hum.SendMsg(&rspMsg)
		return
	}

	t := memberPtr.team
	roomPtr.removeObTeam(t)
	roomPtr.addTeam(t)
	roomPtr.autoStartCountdown()

	hum.SendMsg(&rspMsg)
	roomPtr.notifyRoomInfo(hum)
}

func (this *roomMgr) CMSwitchTeam(hdr *f5.MsgHdr, msg *cs.CMSwitchTeam) {
	hum := hdr.Context.(common.Player)
	rspMsg := cs.SMSwitchTeam{}

	roomPtr, ok := hum.GetRoom().(*room)
	if !ok || roomPtr == nil {
		rspMsg.Errcode = proto.Int32(1)
		rspMsg.Errmsg = proto.String("room is empty")
		hum.SendMsg(&rspMsg)
		return
	}

	if !roomPtr.isOwner(hum) {
		rspMsg.Errcode = proto.Int32(1)
		rspMsg.Errmsg = proto.String("is not room owner")
		hum.SendMsg(&rspMsg)
		return
	}

	if msg.GetType() == 0 {
		t := roomPtr.getObTeamById(msg.GetTeamId())
		if t == nil {
			rspMsg.Errcode = proto.Int32(1)
			rspMsg.Errmsg = proto.String("team_id error")
			hum.SendMsg(&rspMsg)
			return
		}
		if roomPtr.isTeamFull() {
			rspMsg.Errcode = proto.Int32(3)
			rspMsg.Errmsg = proto.String("room team is full")
			hum.SendMsg(&rspMsg)
			return
		}
		roomPtr.removeObTeam(t)
		roomPtr.addTeam(t)
		roomPtr.autoStartCountdown()
	} else {
		t := roomPtr.getTeamById(msg.GetTeamId())
		if t == nil {
			rspMsg.Errcode = proto.Int32(1)
			rspMsg.Errmsg = proto.String("team_id error")
			hum.SendMsg(&rspMsg)
			return
		}
		roomPtr.removeTeam(t)
		roomPtr.addObTeam(t)
	}

	hum.SendMsg(&rspMsg)
	roomPtr.notifyRoomInfo(hum)
}
