package friend

import (
	"q5"
	"f5"
	"fmt"
	"main/constant"
	"main/model"
	. "main/global"
)

type friendMgr struct {
	friendHash map[string]map[string]int32
	blackHash map[string]map[string]int32
}

func (this *friendMgr) Init() {
	this.friendHash = make(map[string]map[string]int32)
	this.blackHash = make(map[string]map[string]int32)
	this.loadFromDB()
}

func (this *friendMgr) UnInit() {
}

func (this *friendMgr) loadFromDB() {
	this.loadFriendships()
	this.loadBlacklist()
}

func (this *friendMgr) loadFriendships() {
	f5.GetSysLog().Info("friendMgr.loadFriendships begin")
	lastIdx := f5.GetJsStyleDb().SyncBatchLoadFullTable(
		constant.FRIEND_DB,
		"SELECT * FROM t_friend_relationship WHERE idx > %d AND deleted = 0",
		func (ds *f5.DataSet) {
			accountId1 := ds.GetByName("account_id1")
			accountId2 := ds.GetByName("account_id2")
			addTime := q5.ToInt32(ds.GetByName("add_time"))
			this.addFriendShip(accountId1, accountId2, addTime)
		},
		func (err error) {
			panic(fmt.Sprintf("friendMgr.loadFriendships dberror:%s", err))
		})
	f5.GetSysLog().Info("friendMgr.loadFriendships end lastIdx:%d friendNum:%d blackNum:%d",
		lastIdx,
		len(this.friendHash),
		len(this.blackHash))
}

func (this *friendMgr) loadBlacklist() {
	f5.GetSysLog().Info("friendMgr.loadBlacklist begin")
	lastIdx := f5.GetJsStyleDb().SyncBatchLoadFullTable(
		constant.FRIEND_DB,
		"SELECT * FROM t_friend_blacklist WHERE idx > %d AND deleted = 0",
		func (ds *f5.DataSet) {
			accountId := ds.GetByName("account_id")
			blockId := ds.GetByName("block_id")
			addTime := q5.ToInt32(ds.GetByName("add_time"))
			this.addBlackList(accountId, blockId, addTime)
		},
		func (err error) {
			panic(fmt.Sprintf("friendMgr.loadBlacklist dberror:%s", err))
		})
	f5.GetSysLog().Info("friendMgr.loadBlacklist end lastIdx:%d friendNum:%d blackNum:%d",
		lastIdx,
		len(this.friendHash),
		len(this.blackHash))
}

func (this *friendMgr) IsFriend(accountId1 string, accountId2 string) bool {
	myFriends := this.getFriends(accountId1)
	if myFriends != nil {
		if _, ok := (*myFriends)[accountId2]; ok {
			return ok
		}
	}
	return false
}

func (this *friendMgr) HasFriend(accountId string) bool {
	return this.GetFriendNum(accountId) > 0
}

func (this *friendMgr) GetFriendNum(accountId string) int32 {
	myFriends := this.getFriends(accountId)
	if myFriends != nil {
		return int32(len(*myFriends))
	}
	return 0
}

func (this *friendMgr) TraverseFriend(accountId string, cb func(string, int32) bool) {
	myFriends := this.getFriends(accountId)
	if myFriends != nil {
		for a, t := range *myFriends {
			if !cb(a, t) {
				break
			}
		}
	}
}

func (this *friendMgr) GetFriendList(accountId string) []string {
	friendList := []string{}
	this.TraverseFriend(
		accountId,
		func (friendId string, addTime int32) bool {
			ele := q5.NewSliceElement(&friendList)
			*ele = friendId
			return true
		})
	return friendList
}

func (this *friendMgr) TraverseBlack(accountId string, cb func(string, int32) bool) {
	myBlacks := this.getBlacks(accountId)
	if myBlacks != nil {
		for a, t := range *myBlacks {
			if !cb(a, t) {
				break
			}
		}
	}
}

func (this *friendMgr) GetBlackList(accountId string) []string {
	blackList := []string{}
	this.TraverseBlack(
		accountId,
		func (blackId string, addTime int32) bool {
			ele := q5.NewSliceElement(&blackList)
			*ele = blackId
			return true
		})
	return blackList
}

func (this *friendMgr) getFriends(accountId string) *map[string]int32 {
	if friends, ok := this.friendHash[accountId]; ok {
		return &friends
	}
	return nil
}

func (this *friendMgr) IsBlack(senderId string, targetId string) bool {
	myBlacks := this.getBlacks(senderId)
	if myBlacks != nil {
		if _, ok := (*myBlacks)[targetId]; ok {
			return ok
		}
	}
	return false
}


func (this *friendMgr) getBlacks(accountId string) *map[string]int32 {
	if blacks, ok := this.blackHash[accountId]; ok {
		return &blacks
	}
	return nil
}

func (this *friendMgr) AsyncGetApplyList(lastIdx int64, accountId string,
	cb func(int32, string, int64, []string)) {
	f5.GetJsStyleDb().PageQuery(
		constant.FRIEND_DB,
		50,
		0,
		"SELECT * FROM t_friend_apply WHERE 1=1",
		[]string{},
		f5.GetDbFilter().Comp(
			f5.GetDbFilter().GT("idx", q5.ToString(lastIdx)).And(),
			f5.GetDbFilter().EQ("target_id", accountId).And(),
			f5.GetDbFilter().EQ("status", q5.ToString(constant.FRIEND_APPLY_STATUS_NONE)),
		),
		"",
		func (err error, pg *f5.Pagination) {
			var lastSinceId int64 = lastIdx
			if err != nil {
				cb(500, "", lastSinceId, []string{})
				return
			}
			users := []string{}
			for pg.Rows.Next() {
				idx := q5.ToInt64(pg.Rows.GetByName("idx"))
				senderId := pg.Rows.GetByName("sender_id")
				if idx > lastSinceId {
					lastSinceId = idx
				} else {
					panic(fmt.Sprintf("AsyncGetApply idx error:%s %s", idx, lastSinceId))
				}
				if this.IsFriend(accountId, senderId) {
					model.FriendApply.SetStatus(senderId, accountId, constant.FRIEND_APPLY_STATUS_ACCEPT,
						func (error, int64, int64) {

						})
				} else {
					*q5.NewSliceElement(&users) = senderId
				}
			}
			cb(0, "", lastSinceId, users)
		})
}

func (this *friendMgr) AsyncApplyFriend(senderId string, targetId string, cb func(int32, string)) {
	if this.IsFriend(senderId, targetId) {
		cb(0, "")
		return
	}
	if this.IsBlack(targetId, senderId) {
		cb(0, "")
		return
	}
	model.FriendApply.Force(senderId, targetId,
		func (err error, lastInsertId int64, rowsAffected int64) {
			if err != nil {
				cb(1, "")
				return
			}
			cb(0, "")
		})
}

func (this *friendMgr) AsyncAcceptApply(senderId string, targetId string, cb func(int32, string)) {
	if senderId == targetId {
		cb(1, "")
		return
	}
	accountId1, accountId2 := this.sortAccounts(senderId, targetId)
	func () {
		if this.IsFriend(senderId, targetId) {
			model.FriendApply.SetStatus(senderId, targetId, constant.FRIEND_APPLY_STATUS_ACCEPT,
				func (error, int64, int64) {

				})
			cb(0, "")
			return
		} else {
			nowTime := f5.GetApp().GetNowSeconds()
			model.Friend.Force(accountId1, accountId2, nowTime,
				func (err error, lastInsertId int64, rowsAffected int64) {
					if err != nil {
						cb(1, "")
						return
					}
					model.FriendApply.SetStatus(senderId, targetId, constant.FRIEND_APPLY_STATUS_ACCEPT,
						func (error, int64, int64) {

						})
					GetDbLogMgr().FriendAdd(senderId, targetId)
					this.addFriendShip(accountId1, accountId2, int32(nowTime))
					cb(0, "")
				})
		}
	}()
}

func (this *friendMgr) AsyncRejectApply(senderId string, targetId string, cb func(int32, string)) {
	if senderId == targetId {
		cb(1, "")
		return
	}
	model.FriendApply.SetStatus(senderId, targetId, constant.FRIEND_APPLY_STATUS_REJECT,
		func (error, int64, int64) {

		})
	cb(0, "")
}

func (this *friendMgr) AsyncDeleteFriend(senderId string, targetId string, cb func(int32, string)) {
	if !this.IsFriend(senderId, targetId) ||
		senderId == targetId {
		cb(0, "")
		return
	}
	accountId1, accountId2 := this.sortAccounts(senderId, targetId)
	model.Friend.DeleteSoft(accountId1, accountId2,
		func (err error, lastInsertId int64, rowsAffected int64) {
			if err != nil {
				cb(1, "")
				return
			}
			GetDbLogMgr().FriendDel(senderId, targetId)
			this.removeFriendShip(accountId1, accountId2)
			cb(0, "")
		})
}

func (this *friendMgr) AsyncAddBlack(senderId string, targetId string, cb func(int32, string)) {
	if this.IsBlack(senderId, targetId) {
		cb(0, "")
		return
	}
	if senderId == targetId {
		cb(1, "")
		return
	}
	nowTime := f5.GetApp().GetNowSeconds()
	model.FriendBlack.Force(senderId, targetId, nowTime,
		func (err error, lastInsertId int64, rowsAffected int64) {
			if err != nil {
				cb(1, "")
				return
			}
			this.addBlackList(senderId, targetId, int32(nowTime))
			cb(0, "")
		})
}

func (this *friendMgr) AsyncRemoveBlack(senderId string, targetId string, cb func(int32, string)) {
	if !this.IsBlack(senderId, targetId) {
		cb(0, "")
		return
	}
	model.FriendBlack.DeleteSoft(senderId, targetId,
		func (err error, lastInsertId int64, rowsAffected int64) {
			if err != nil {
				cb(1, "")
				return
			}
			this.removeBlackList(senderId, targetId)
			cb(0, "")

		})
}

func (this *friendMgr) AsyncHasApply(accountId string, cb func(int32, string, int32)) {
	f5.GetJsStyleDb().OrmSelectOne(
		constant.FRIEND_DB,
		"t_friend_apply",
		[][]string{
			{"target_id", accountId},
			{"status", q5.ToString(constant.FRIEND_APPLY_STATUS_NONE)},
		},
		func (err error, ds *f5.DataSet) {
			if err != nil {
				cb(500, "", 0)
				return
			}
			if ds.Next() {
				cb(0, "", 1)
			} else {
				cb(0, "", 0)
			}
		})
}

func (this *friendMgr) addFriendShip(accountId1 string, accountId2 string, addTime int32) {
	{
		friends := this.getFriends(accountId1)
		if friends == nil {
			this.friendHash[accountId1] = make(map[string]int32)
			friends = this.getFriends(accountId1)
		}
		(*friends)[accountId2] = addTime
	}
	{
		friends := this.getFriends(accountId2)
		if friends == nil {
			this.friendHash[accountId2] = make(map[string]int32)
			friends = this.getFriends(accountId2)
		}
		(*friends)[accountId1] = addTime
	}
}

func (this *friendMgr) removeFriendShip(accountId1 string, accountId2 string) {
	{
		friends := this.getFriends(accountId1)
		if friends != nil {
			delete(*friends, accountId2)
		}
	}
	{
		friends := this.getFriends(accountId2)
		if friends != nil {
			delete(*friends, accountId1)
		}
	}
}

func (this *friendMgr) addBlackList(accountId string, blockId string, addTime int32) {
	{
		blacks := this.getBlacks(accountId)
		if blacks == nil {
			this.blackHash[accountId] = make(map[string]int32)
			blacks = this.getBlacks(accountId)
		}
		(*blacks)[blockId] = addTime
	}
}

func (this *friendMgr) removeBlackList(accountId string, blockId string) {
	{
		blacks := this.getBlacks(accountId)
		if blacks != nil {
			delete(*blacks, blockId)
		}
	}
}

func (this *friendMgr) sortAccounts(senderId string, targetId string) (string, string) {
	if senderId < targetId {
		return senderId, targetId
	} else {
		return targetId, senderId
	}
}
