package guild

import (
	"q5"
	"f5"
	"cs"
	"github.com/golang/protobuf/proto"
	"main/constant"
	"main/model"
	. "main/global"
)

type guild struct {
	guildId string
	guildName string
	creatorId string
	badge int32
	notice string
	joinCondType int32
	joinCondVal int32
	maxMemberNum int32
	createTime int32
	modifyTime int32
	totalStars int32
	idHash map[string]*member
}

func (this *guild) GetGuildId() string {
	return this.guildId
}

func (this *guild) GetGuildName() string {
	return this.guildName
}

func (this *guild) GetMemberNum() int32 {
	return int32(len(this.idHash))
}

func (this *guild) loadFromDb(ds *f5.DataSet) {
	this.guildId = ds.GetByName("guild_id")
	this.guildName = ds.GetByName("guild_name")
	this.creatorId = ds.GetByName("creator_id")
	this.badge = q5.ToInt32(ds.GetByName("badge"))
	this.notice = ds.GetByName("notice")
	this.joinCondType = q5.ToInt32(ds.GetByName("join_cond_type"))
	this.joinCondVal = q5.ToInt32(ds.GetByName("join_cond_val"))
	this.maxMemberNum = q5.ToInt32(ds.GetByName("max_members"))
	this.createTime = q5.ToInt32(ds.GetByName("createtime"))
	this.modifyTime = q5.ToInt32(ds.GetByName("modifytime"))
}

func (this *guild) init(guildId string, guildName string, creatorId string, badge int32,
	maxMemberNum int32, nowTime int32) {
	this.guildId = guildId
	this.guildName = guildName
	this.creatorId = creatorId
	this.badge = badge
	this.maxMemberNum = maxMemberNum
	this.createTime = nowTime
	this.modifyTime = nowTime
}

func (this *guild) addMember(m *member) {
	this.idHash[m.memberId] = m
}

func (this *guild) getMember(accountId string) *member {
	if m, ok := this.idHash[accountId]; ok {
		return m
	}
	return nil
}

func (this *guild) traverseMembers(cb func(*member) bool) {
	for _, m := range(this.idHash) {
		if !cb(m) {
			break
		}
	}
}

func (this *guild) isOwner(accountId string) bool {
	if m := this.getMember(accountId); m != nil {
		return m.guildJob == constant.GuildMemberLevelLeader
	}
	return false
}

func (this *guild) updateByKv(kv map[int32]string) {
	for k, v := range(kv) {
		if k == constant.GUILD_UPDATE_FIELD_NOTICE {
			this.notice = v
		} else if k == constant.GUILD_UPDATE_FIELD_AVATAR {
			this.badge = q5.ToInt32(v)
		} else if k == constant.GUILD_UPDATE_FIELD_NAME {
			this.guildName = v
		} else if k == constant.GUILD_UPDATE_FIELD_COND_TYPE {
			this.joinCondType = q5.ToInt32(v)
		} else if k == constant.GUILD_UPDATE_FIELD_COND_VALUE {
			this.joinCondVal = q5.ToInt32(v)
		}
	}
}

func (this *guild) disband(ownerId string, reason int32)  {
	model.Guild.DeleteSoft(this.guildId,
		func (err error, lastInsertId int64, rowsAffected int64) {
		})
	model.GuildApply.Clear(this.guildId,
		func (err error, lastInsertId int64, rowsAffected int64) {
		})
	_guildMgr.delGuild(this)
	GetDbLogMgr().GuildDisband(this.guildId, ownerId, reason)
}

func (this *guild) AsyncFillMFGuild(pbGuild *cs.MFGuild, cb func(int32, string)) {
	pbGuild.AutoId = proto.Int64(0)
	pbGuild.GuildId = proto.Int64(q5.ToInt64(this.guildId))
	pbGuild.Name = proto.String(this.guildName)
	pbGuild.LeaderId = proto.String(this.getOwnerId())
	pbGuild.Avatar = proto.Int32(this.badge)
	pbGuild.Notice = proto.String(this.notice)
	pbGuild.JoinCond = proto.Int32(this.joinCondType)
	pbGuild.JoinCondValue = proto.Int32(this.joinCondVal)
	pbGuild.TotalStars = proto.Int32(0)
	pbGuild.TotalKills = proto.Int32(0)
	pbGuild.ChickenDinners = proto.Int32(0)
	pbGuild.MaxMembers = proto.Int32(this.maxMemberNum)
	accountIds := []string{}
	for _, m := range(this.idHash) {
		q5.AppendSlice(&accountIds, m.memberId)
	}
	q5.NewSlice(&pbGuild.Members, 0, int32(len(accountIds)))
	GetCacheMgr().AsyncGetUsersAndFillMFGuildMember(
		accountIds,
		&pbGuild.Members,
		func (errCode int32, errMsg string) {
			if errCode != 0 {
				cb(500, "server internal error")
				return
			}
			for _, m := range(pbGuild.Members) {
				if m.GetAccountId() == pbGuild.GetLeaderId() {
					pbGuild.LeaderName = proto.String(m.GetUsername())
					break
				}
			}
			cb(0, "")
		})
}

func (this *guild) asyncLeave(accountId string, cb func(int32, string)) {
	m := this.getMember(accountId)
	if m == nil {
		cb(0, "")
		return
	}
	model.GuildMember.DeleteSoft(this.guildId, m.memberId,
		func (err error, lastInsertId int64, rowsAffected int64) {
			if err != nil {
				cb(500, "server internal error")
				return
			}
			this.delMember(m.memberId)
			_guildMgr.delGuildMember(m.memberId)
			GetDbLogMgr().GuildLeave(this.guildId, m.memberId)
			if m.isJob(constant.GuildMemberLevelLeader) {
				nextLeader := this.chooseLeader()
				if nextLeader != nil {
					model.GuildMember.SetJob(this.guildId, m.memberId, constant.GuildMemberLevelLeader,
						func (err error, lastInsertId int64, rowsAffected int64) {
							if err != nil {
								cb(500, "server internal error")
								return
							}
							nextLeader.guildJob = constant.GuildMemberLevelLeader
							cb(0, "")
							return
						})
				} else {
					if this.GetMemberNum() <= 0 {
						this.disband(accountId, 0)
					}
					cb(0, "")
					return
				}
			} else {
				if this.GetMemberNum() <= 0 {
					this.disband(accountId, 0)
				}
				cb(0, "")
			}
		})
}

func (this *guild) getOwnerId() string {
	ownerId := ""
	this.traverseMembers(
		func (m *member) bool {
			if m.guildJob == constant.GuildMemberLevelLeader {
				ownerId = m.memberId
				return false
			}
			return true
		})
	return ownerId
}

func (this *guild) chooseLeader() *member {
	var nextLeader *member
	this.traverseMembers(
		func (ele *member) bool {
			if nextLeader == nil {
				nextLeader = ele
			} else if ele.guildJob < nextLeader.guildJob {
				nextLeader = ele
			}
			return true
		})
	return nextLeader
}

func (this *guild) delMember(memberId string) {
	delete(this.idHash, memberId)
}

func (this *guild) isFull() bool {
	return this.GetMemberNum() >= this.maxMemberNum
}

func newGuild() *guild {
	p := new(guild)
	p.idHash = make(map[string]*member)
	return p
}
